program es_blas
implicit none

integer :: i, j
character(len=*) , parameter :: FMT='(12E12.5)'
! BLAS routines and funcions
external :: dcopy, dscal, dnrm2, ddot
external :: dgemv, dger, dgemm
real*8 :: dnrm2, ddot
! Matrix
integer :: M, N, K
parameter ( M=12, N=8, K=4 )
real*8 , allocatable :: A(:,:), B(:,:), C(:,:), D(:,:)
real*8 , allocatable :: X(:), Y(:), Z(:), V(:)
! Memory allocation
allocate( A( M, K ) )
allocate( B( K, N ) )
allocate( C( M, N ) )
allocate( D( K, N ) )
allocate( X( M ) )
allocate( Y( N ) )
allocate( Z( N ) )
allocate( V( K ) )

! Vector and matrix elements generation
do i=1,N
  X(i) = dble(i)
  Y(i) = dble(N+1-i)
enddo
do i=N+1,M
  X(i) = dble(i)
enddo
write(*,*) 'VECTOR X:'
write(*,FMT) ( X(i), i=1,M )
write(*,*) 'VECTOR Y:'
write(*,FMT) ( Y(i), i=1,N )
do j=1,K
  do i=1,M
    A(i,j) = dble(K*(i-1)+j)
  enddo
enddo
write(*,*) 'MATRIX A:'
do i=1,M
   write(*,FMT) ( A(i,j), j=1,K )
enddo
do j=1,N
  do i=1,K
    B(i,j) = dble(N*(K-i+1)-j+1)
  enddo
enddo
write(*,*) 'MATRIX B:'
do i=1,K
   write(*,FMT) ( B(i,j), j=1,N )
enddo

!           N, DX, INCX, DY, INCY
call dcopy( N,  Y,    1,  Z,    1 )
write(*,*) '### DCOPY ###'
write(*,*) 'VECTOR Z:'
write(*,FMT) ( Z(i), i=1,N )

!           N,    DA, DX, INCX
call dscal( N, 2.0d0,  Z,    1 )
write(*,*) '### DSCAL ###'
write(*,*) 'VECTOR Z:'
write(*,FMT) ( Z(i), i=1,N )

write(*,*) '### DNRM2 ###'
!                                     N, X, INCX
write(*,*) 'sqrt( Z''*Z ) = ', dnrm2( N, Z,    1 )

write(*,*) '### DDOT ###'
!                              N, DX, INCX, DY, INCY
write(*,*) 'Z dot Y = ', ddot( N,  Z,    1,  Y,    1 )

!          TRANS, M, N, ALPHA, A, LDA, X, INCX,  BETA, Y, INCY
call dgemv(  'N', K, N, 1.0d0, B,   K, Y,    1, 0.0d0, V,    1 )
write(*,*) '### DGEMV ###'
write(*,*) 'VECTOR V:'
write(*,FMT) ( V(i), i=1,K )

!          M, N, ALPHA, X, INCX, Y, INCY, A, LDA
call dger( K ,N, 1.0d0, V,    1, Y,    1, D,   K )
write(*,*) '### DGER ###'
write(*,*) 'MATRIX D:'
do i=1,K
  write(*,FMT) ( D(i,j), j=1,N )
enddo

!          TRANSA, TRANSB, M, N, K, ALPHA, A, LDA, B, LDB,  BETA, C, LDC
call dgemm(   'N',    'N', M, N, K, 1.0d0, A,   M, B,   K, 0.0d0, C,   M )
write(*,*) '### DGEMM ###'
write(*,*) 'MATRIX C:'
do i=1,M
  write(*,FMT) ( C(i,j), j=1,N )
enddo

end program es_blas

