program es_scalapack
use mpi
implicit none

integer :: i, j
! MPI
integer :: ierr, nprocs, myrank
! BLACS/SCALAPACK routines and functions
external :: blacs_exit, blacs_gridexit, blacs_gridinfo, descinit, numroc
external :: pdelset, pdelget, pdlawrite, pdlaread
integer :: numroc
integer :: context, nprow, npcol, iam, myrow, mycol, info
! PBLAS routines and funcions
external :: pdcopy, pdscal, pdnrm2, pddot
external :: pdgemv, pdger, pdgemm

! Vectors and Matrix
integer :: M, N, K, MB, NB, KB, Nloc, Mloc, KRloc, KCloc, NVloc
parameter ( M=8, N=8, K=8, MB=2, NB=2, KB=2 )
real*8 , allocatable :: A(:,:), B(:,:), C(:,:), D(:,:)
real*8 , allocatable :: X(:), Y(:), Z(:), V(:), work(:)
real*8 :: NRM2, DOT
integer :: descA(9), descB(9), descC(9), descD(9)
integer :: descX(9), descY(9), descZ(9), descV(9)

! Initialize MPI environment
call MPI_Init(ierr)
call MPI_Comm_Size(mpi_comm_world,nprocs,ierr)
call MPI_Comm_Rank(mpi_comm_world,myrank,ierr)

! Initialize a BLACS context
call sl_init_rm( iam, context, nprow, npcol )
call blacs_gridinfo( context, nprow, npcol, myrow, mycol )

! Computation of local matrix size
Mloc  = numroc( M, MB, myrow, 0, nprow )
KCloc = numroc( K, KB, mycol, 0, npcol )
KRloc = numroc( K, KB, myrow, 0, nprow )
Nloc  = numroc( N, NB, mycol, 0, npcol )
NVloc = numroc( N, NB, myrow, 0, nprow )
allocate( A( Mloc, KCloc ) )
allocate( B( KRloc, Nloc ) )
allocate( C( Mloc, Nloc ) )
allocate( D( Mloc, Nloc ) )
allocate( X( Mloc ) )
allocate( Y( NVloc ) )
allocate( Z( NVloc ) )
allocate( V( Mloc ) )
allocate( work( M ) )
! Descriptors
call descinit(descA, M, K, MB, KB, 0, 0, context, max(1,Mloc),  info)
call descinit(descB, K, N, KB, NB, 0, 0, context, max(1,KRloc), info)
call descinit(descC, M, N, MB, NB, 0, 0, context, max(1,Mloc),  info)
call descinit(descD, M, N, MB, NB, 0, 0, context, max(1,Mloc),  info)
call descinit(descX, M, 1, MB,  1, 0, 0, context, max(1,Mloc),  info)
call descinit(descY, N, 1, NB,  1, 0, 0, context, max(1,NVloc), info)
call descinit(descZ, N, 1, NB,  1, 0, 0, context, max(1,NVloc), info)
call descinit(descV, M, 1, MB,  1, 0, 0, context, max(1,Mloc),  info)
call MPI_Barrier( mpi_comm_world, ierr)

! Vector and matrix elements generation
do i=1,N
  call pdelset( X, i, 1, descX, dble(i) )
  call pdelset( Y, i, 1, descY, dble(N+1-i) )
enddo
do i=N+1,M
  call pdelset( X, i, 1, descX, dble(i) )
enddo
call pdlawrite('x.dat', M, 1, X, 1, 1, descX, 0, 0, work)
call pdlawrite('y.dat', N, 1, Y, 1, 1, descY, 0, 0, work)
do j=1,K
  do i=1,M
    call pdelset( A, i, j, descA, dble(K*(i-1)+j) )
  enddo
enddo
call pdlawrite('A.dat', M, K, A, 1, 1, descA, 0, 0, work)
do j=1,N
  do i=1,K
    call pdelset( B, i, j, descB, dble(N*(K-i+1)-j+1) )
  enddo
enddo
call pdlawrite('B.dat', K, N, B, 1, 1, descB, 0, 0, work)

!            N, X, IX, JX, DESCX, INCX, Y, IY, JY, DESCY, INCY
call pdcopy( N, Y,  1,  1, descY,    1, Z,  1,  1, descZ,    1  )
call pdlawrite('z_copy.dat', N, 1, Z, 1, 1, descZ, 0, 0, work)

!            N, ALPHA, X, IX, JX, DESCX, INCX
call pdscal( N, 2.0d0, Z,  1,  1, descZ,    1 )
call pdlawrite('z_scal.dat', N, 1, Z, 1, 1, descZ, 0, 0, work)

!            N, NRM2, X, IX, JX, DESCX, INCX
call pdnrm2( N, NRM2, Z,  1,  1, descZ,    1  )
if(myrank.eq.0) then
  write(*,*) '### PDNRM2 ###'
  write(*,*) 'sqrt( Z''*Z ) = ', NRM2, myrank
endif

!           N, DOT, X, IX, JX, DESCX, INCX, Y, IY, JY, DESCY, INCY
call pddot( N, DOT, Z,  1,  1, descZ,    1, Y,  1,  1, descY,    1  )
if(myrank.eq.0) then
  write(*,*) '### PDDOT ###'
  write(*,*) 'Z dot Y = ', DOT, myrank
endif

!           TRANS, M, N, ALPHA, A, IA, JA, DESCA, X, IX, JX, DESCX, INCX,  BETA, Y, IY, JY, DESCY, INCY
call pdgemv(  'N', M, K, 1.0d0, A,  1,  1, descA, X,  1,  1, descX,    1, 0.0d0, V,  1,  1, descV,    1  )
call pdlawrite('v_pdgemv.dat', M, 1, V, 1, 1, descV, 0, 0, work)

!           M,  N, ALPHA, X, IX, JX, DESCX, INCX, Y, IY, JY, DESCY, INCY, A, IA, JA, DESCA
call pdger( M , N, 1.0d0, X,  1,  1, descX,    1, Y,  1,  1, descY,    1, C,  1,  1, descC )
call pdlawrite('C_pdger.dat', M, N, C, 1, 1, descC, 0, 0, work)

!          TRANSA, TRANSB, M, N, K, ALPHA, A, IA, JA, DESCA, B, IB, YB, DESCB,  BETA, C, IC, JC, DESCC
call dgemm(   'N',    'N', M, N, K, 1.0d0, A,  1,  1, descA, B,  1,  1, descB, 0.0d0, D,  1,  1, descD )
call pdlawrite('D_pdgemm.dat', M, N, D, 1, 1, descD, 0, 0, work)

call MPI_Barrier( mpi_comm_world, ierr)

deallocate( A )
deallocate( B )
deallocate( C )
deallocate( D )
deallocate( X )
deallocate( Y )
deallocate( Z )
deallocate( V )
deallocate( work )

! Close BLACS environment
call blacs_gridexit( context )
call blacs_exit( 0 )

! Close MPI environment
!call MPI_Finalize(ierr)

end program es_scalapack

!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!!
!  file: SL_INIT_RM.f
      SUBROUTINE SL_INIT_RM( IAM, ICTXT, NPROW, NPCOL )
!                            OUT   OUT    OUT    OUT
!
!     .. Scalar Arguments ..
      INTEGER IAM, ICTXT, NPCOL, NPROW
!     ..
!
!  Purpose
!  =======
!
!  SL_INIT_RM initializes an NPROW x NPCOL process grid using a row-major
!  ordering (RM)  of  the  processes. This routine retrieves a default system
!  context which will include all available processes.
!
!  Arguments
!  =========
!
!  IAM     (global output) INTEGER
!          IAM is the rank of the procces
!
!  ICTXT   (global output) INTEGER
!          ICTXT specifies the BLACS context handle identifying the
!          created process grid.  The context itself is global.
!
!  NPROW   (global input) INTEGER
!          NPROW specifies the number of process rows in the grid
!          that will be created.  
!
!  NPCOL   (global input) INTEGER
!          NPCOL specifies the number of process columns in the grid
!          that will be created.
!
!  =====================================================================
!
!     .. Local Scalars ..
      INTEGER NPROC
!     ..
!     .. External Subroutines ..
      EXTERNAL BLACS_GET, BLACS_GRIDINIT, BLACS_PINFO
!     ..
!     .. Executable Statements ..
!
!     Get starting information
!
      CALL BLACS_PINFO( IAM, NPROC )
!                       OUT   OUT
!
! -----    Set the dimension of the 2d processors grid.
!
      call gridsetup(NPROC,NPROW,NPCOL)
!                      IN   OUT   OUT
!
!     Define process grid
!
      CALL BLACS_GET( -1, 0, ICTXT )
!                     IN IN   OUT
      CALL BLACS_GRIDINIT( ICTXT, 'Row-major', NPROW, NPCOL )
!                            IN     IN           IN     IN
!
      RETURN
!
!     End of SL_INIT_RM
!
      END

!-----------------------------------------------------------------------
!
      subroutine gridsetup(nproc,nprow,npcol)
!
! This subroutine factorizes the number of processors (nproc)
! into nprow and npcol, that are the sizes of the 2d processors mesh.
!
      implicit none
      integer nproc,nprow,npcol
      integer sqrtnp,i 

      sqrtnp = int( sqrt( dble(nproc) ) + 1 )
      do i=1,sqrtnp
        if(mod(nproc,i).eq.0) nprow = i
      end do
      npcol = nproc/nprow

      return
      end

